/**
  *  \file lispy.cc
  *  \brief Not Quite Lisp Nor Scheme
  *
  *  See main() function at the end for what Lisp functions are
  *  implemented.
  *
  *  This is experimental code and by no means complete.
  *
  *  (c) 2005-2007 Stefan Reuther <Streu@gmx.de>
  */

#include <cctype>
#include <cstdlib>
#include <map>
#include <cstdio>
#include <iostream>
#include <list>
#include <climits>
#include <fstream>
#include <sstream>
#include "lispy.h"

static LPtr SFQuote(LPtr p);

LObject::~LObject()
{ }

/** Compare for equality.
    \param other      object to compare to
    \param identical  true iff objects should be identical,
                      false iff objects should be equal component-wise

    In Emacs Lisp, equal literal integers or symbols compare
    identical, but equal string literals or floats do not. We
    currently clone that behaviour. Common Lisp allows that equal
    integers compare non-identical, but symbols must always be
    equal. */
bool
LObject::equals(LPtr other, bool /*identical*/)
{
    return other.operator->() == this; // FIXME: hack
}

/******************************* LSelfEval *******************************/

LSelfEval::~LSelfEval()
{ }

LPtr
LSelfEval::eval()
{
    return this;
}

void
LSelfEval::setf(LPtr /*val*/)
{
    throw LLispException("invalid-setf", "Cannot set this value");
}

/********************************* LSelfEvalSym *********************************/

LSelfEvalSym::LSelfEvalSym(const char* name)
    : name(name)
{
    ref = 1; /* avoid that someone attempts to free it */
}

std::string
LSelfEvalSym::toString()
{
    return name;
}

//- nil: means false.
LSelfEvalSym sym_nil("nil");
//- t: means true. Actually, everything which is not nil means true.
LSelfEvalSym sym_t("t");

/********************************** LPtr *********************************/

LPtr::LPtr()
    : p(&sym_nil)
{
    ++p->ref;
}

LPtr::LPtr(const LPtr& a)
    : p(a.p)
{
    ++p->ref;
}

LPtr::LPtr(LObject* p)
    : p(p)
{
    ++p->ref;
}

LPtr::~LPtr()
{
    if (!--p->ref)
        delete p;
}

LPtr&
LPtr::operator=(const LPtr& a)
{
    LObject* other = a.p;
    ++other->ref;
    if (!--p->ref)
        delete p;
    p = other;
    return *this;
}

/********************************** LInt *********************************/

LInt::~LInt()
{ }

std::string
LInt::toString()
{
    char tmp[30];
    std::sprintf(tmp, "%d", n);
    return tmp;
}

bool
LInt::equals(LPtr other, bool /*identical*/)
{
    LInt* p = other.as<LInt>();
    return p && n == p->n;
}

/******************************** LString ********************************/

LString::~LString()
{ }

std::string
LString::toString()
{
    std::string rv = s;
    for (std::string::size_type i = rv.size(); i > 0; --i) {
        if (rv[i-1] == '\\' || rv[i-1] == '"') {
            rv.insert(i-1, "\\");
        } else if (rv[i-1] == '\r') {
            rv.replace(i-1, 1, "\\r");
        } else if (rv[i-1] == '\n') {
            rv.replace(i-1, 1, "\\n");
        } else if (rv[i-1] == '\t') {
            rv.replace(i-1, 1, "\\t");
        }
    }
    return "\"" + rv + "\"";
}

bool
LString::equals(LPtr other, bool identical)
{
    LString* p = other.as<LString>();
    if (!p)
        return false;
    else if (identical)
        return p == this;
    else
        return p->s == this->s;
}


/******************************** LSymbol ********************************/

std::map<std::string,LPtr> symtab;

LSymbol::~LSymbol()
{ }

std::string
LSymbol::toString()
{
    return s;
}

LPtr
LSymbol::eval()
{
    if (s.size() && s[0] == ':') {
        return this;
    } else {
        std::map<std::string,LPtr>::const_iterator i = symtab.find(s);
        if (i == symtab.end())
            throw LLispException("undefined-symbol", "Symbol '" + s + "' is not defined");
        return i->second;
    }
}

void
LSymbol::setf(LPtr p)
{
    if (s.size() && s[0] == ':') {
        throw LLispException("invalid-setf", "Cannot set '" + s + "'");
    } else {
        symtab[s] = p->eval();
    }
}

void
LSymbol::define(std::string name, LPtr value)
{
    symtab[name] = value;
    if (LApplyable* a = value.as<LApplyable>())
        a->setNameMaybe(name);
}

void
LSymbol::defineSetf(std::string get, std::string set)
{
    std::map<std::string,LPtr>::iterator g = symtab.find(get);
    std::map<std::string,LPtr>::iterator s = symtab.find(set);
    if (g == symtab.end() || s == symtab.end())
        throw LLispException("init-error", "define-setf: undefined");

    LApplyable* a = g->second.as<LApplyable>();
    if (!a)
        throw LLispException("init-error", "define-setf: not a function");
    a->defineSet(s->second);
}

bool
LSymbol::equals(LPtr other, bool /*identical*/)
{
    LSymbol* p = other.as<LSymbol>();
    return p && p->s == s;
}

/********************************* Utils *********************************/

bool
isList(LPtr p)
{
    while (LCons* c = p.as<LCons>())
        p = c->getCdr();
    return p == &sym_nil;
}

int
getListLength(LPtr p)
{
    int n = 0;
    while (LCons* c = p.as<LCons>())
        p = c->getCdr(), ++n;
    if (!p)
        return n;
    throw LLispException("malformed-list", "Malformed list");
}

LPtr
getCar(LPtr p)
{
    if (LCons* c = p.as<LCons>())
        return c->getCar();
    throw LLispException("not-cons", "Not cons");
}

LPtr
getCdr(LPtr p)
{
    if (LCons* c = p.as<LCons>())
        return c->getCdr();
    throw LLispException("not-cons", "Not cons");
}

LPtr
getTail(LPtr p, int n)
{
    while (n-- > 0)
        p = getCdr(p);
    return p;
}

LPtr
getNth(LPtr p, int n)
{
    return getCar(getTail(p, n));
}

LPtr
evalList(LPtr p)
{
    LPtr rv;
    while (LCons* c = p.as<LCons>()) {
        rv = c->getCar()->eval();
        p = c->getCdr();
    }
    if (!!p)
        throw LLispException("malformed-list", "Malformed list");
    return rv;
}

LPtr
nconc(LPtr list, LPtr last)
{
    if (!list) {
        return last;
    } else {
        LPtr p = list;
        while (LCons* c = p.as<LCons>()) {
            if (!c->getCdr()) {
                c->setCdr(last);
                return list;
            } else {
                p = c->getCdr();
            }
        }
        throw LLispException("malformed-list", "Malformed list");
    }
}

LPtr
appendByCopying(LPtr list, LPtr nlast)
{
    LPtr n;
    LCons root(n, n);
    LCons* last = &root;
    while (LCons* c = list.as<LCons>()) {
        LPtr val = c->getCar();
        LCons* nc = new LCons(val, LPtr());
        last->setCdr(nc);
        last = nc;
        list = c->getCdr();
    }
    last->setCdr(nlast);
    return root.getCdr();
}

/********************************* LCons *********************************/

LCons::~LCons()
{ }

LPtr
LCons::eval()
{
    LPtr lhs = car->eval();
    if (LApplyable* a = lhs.as<LApplyable>()) {
        if (!isList(cdr))
            throw LLispException("argument-syntax", "Argument syntax error");
        return a->apply(cdr);
    } else {
        throw LLispException("invalid-function", "Invalid function '" + lhs->toString() + "'");
    }
}

void
LCons::setf(LPtr val)
{
    LPtr lhs = car->eval();
    if (LApplyable* a = lhs.as<LApplyable>()) {
        if (!isList(cdr))
            throw LLispException("argument-syntax", "Argument syntax error");
        a->applySet(cdr, val);
    } else {
        throw LLispException("invalid-function", "Invalid function '" + lhs->toString() + "'");
    }
}

std::string
LCons::toString()
{
    std::string accum = "(" + car->toString();
    LPtr p = cdr;
    while (LCons* c = p.as<LCons>()) {
        accum += " ";
        accum += c->getCar()->toString();
        p = c->getCdr();
    }
    if (!p) {
        accum += ")";
    } else {
        accum += " . ";
        accum += p->toString();
        accum += ")";
    }
    return accum;
}

bool
LCons::equals(LPtr other, bool identical)
{
    LCons* p = other.as<LCons>();
    if (!p)
        return false;
    else if (identical)
        return p == this;
    else
        return car->equals(p->getCar(), false)
            && cdr->equals(p->getCdr(), false);
}

/******************************* LApplyable ******************************/

void
LApplyable::setNameMaybe(std::string name)
{
    if (!this->name.size())
        this->name = name;
}

std::string
LApplyable::getName() const
{
    if (this->name.size())
        return this->name;
    else
        return "<unnamed>";
}

void
LApplyable::applySet(LPtr args, LPtr val)
{
    if (!setf_function)
        throw LLispException("invalid-setf", "No setf function defined for '" + getName() + "'");

    args = appendByCopying(args, new LCons(val, LPtr()));
    LApplyable* a = setf_function.as<LApplyable>();
    if (!a)
        throw LLispException("invalid-setf", "Invalid setf function defined for '" + getName() + "'");
    a->apply(args);
}

void
LApplyable::defineSet(LPtr exp)
{
    setf_function = exp;
}

void
LApplyable::setf(LPtr /*val*/)
{
    throw LLispException("invalid-setf", "Cannot setf a function");
}

/******************************** LSpecial *******************************/

LSpecial::~LSpecial()
{ }

LPtr
LSpecial::apply(LPtr p)
{
    return fun(p);
}

std::string
LSpecial::toString()
{
    return "#<special:" + getName() + ">";
}


/****************************** Environment ******************************/

/** Helper class to manage local bindings. */
class Environment {
    std::map<std::string,LPtr> orig_values;
    std::map<std::string,bool> orig_bound;
    std::list<std::string> pending_names;
    std::list<LPtr> pending_values;
 public:
    Environment();
    ~Environment();

    void set(std::string name, LPtr value);
    void bind();
};

Environment::Environment()
{ }

Environment::~Environment()
{
    for (std::map<std::string,bool>::iterator i = orig_bound.begin(); i != orig_bound.end(); ++i) {
        if (i->second)
            symtab[i->first] = orig_values[i->first];
        else
            symtab.erase(i->first);
    }
}

/** Prepare for setting a variable. Does not actually set the variable. */
void
Environment::set(std::string name, LPtr value)
{
    pending_names.push_back(name);
    pending_values.push_back(value);
}

/** Perform all pending variable binds. */
void
Environment::bind()
{
    while (!pending_names.empty()) {
        const std::string& name = pending_names.front();
        const LPtr& value = pending_values.front();

        if (orig_bound.find(name) == orig_bound.end()) {
            std::map<std::string,LPtr>::iterator i = symtab.find(name);
            if (i != symtab.end()) {
                orig_values[name] = i->second;
                orig_bound[name] = true;
            } else {
                orig_bound[name] = false;
            }
        }
        symtab[name] = value;

        pending_names.pop_front();
        pending_values.pop_front();
    }
}

/***************************** LApplyByValue *****************************/

LPtr
LApplyByValue::apply(LPtr p)
{
    // evaluate all args, and call applyEvaluated()
    LPtr n;
    LCons root(n, n);
    LCons* last = &root;
    while (LCons* c = p.as<LCons>()) {
        LPtr val = c->getCar()->eval();
        LCons* nc = new LCons(val, LPtr());
        last->setCdr(nc);
        last = nc;
        p = c->getCdr();
    }
    return applyEvaluated(root.getCdr());
}

/******************************* LFunction *******************************/

LFunction::LFunction(LPtr args, LPtr body)
    : args(args), body(body)
{ }

LFunction::~LFunction()
{ }

LPtr
LFunction::applyEvaluated(LPtr p)
{
    LPtr a = args;
    LCons *ac, *pc;
    Environment env;
    while (((ac = a.as<LCons>()) != 0) & ((pc = p.as<LCons>()) != 0)) {
        LSymbol* arg_name = ac->getCar().as<LSymbol>();
        if (!arg_name)
            throw LLispException("invalid-parameter", "Invalid parameter name");
        env.set(arg_name->getSymbol(), pc->getCar());
        a = ac->getCdr();
        p = pc->getCdr();
    }
    if (!!ac)
        throw LLispException("argument-count", "Too few arguments to '" + getName() + "'");
    if (!!pc)
        throw LLispException("argument-count", "Too many arguments to '" + getName() + "'");
    env.bind();
    try {
        return evalList(body);
    }
    catch (LLispException& e) {
        e.pushTrace(this);
        throw;
    }
}

/***************************** LLispException ****************************/

void
LLispException::pushTrace(LPtr obj)
{
    trace = new LCons(obj, trace);
}

/******************************** LBuiltin *******************************/

LBuiltin::~LBuiltin()
{ }

LPtr
LBuiltin::applyEvaluated(LPtr p)
{
    // validate argument count
    int n = getListLength(p);
    if (n < min)
        throw LLispException("argument-count", "Too few arguments to '" + getName() + "'");
    if (n > max)
        throw LLispException("argument-count", "Too many arguments to '" + getName() + "'");
    try {
        return fun(p);
    }
    catch (LLispException& e) {
        e.pushTrace(this);
        throw;
    }
}

std::string
LBuiltin::toString()
{
    return "#<builtin:" + getName() + ">";
}

/********************************* Parser ********************************/

static void
skipWhitespace(std::istream& in)
{
    using namespace std;
    char c;
    while (in.get(c)) {
        if (c == ';') {
            while (in.get(c) && c != '\n')
                ;
        } else if (!isspace(c)) {
            in.putback(c);
            break;
        }
    }
}

static LPtr
parseTail(std::istream& in)
{
    skipWhitespace(in);

    char c;
    if (!in.get(c))
        throw LLispException("read-syntax", "End of file in list");

    if (c == ')') {
        return LPtr();
    } else if (c == '.') {
        LPtr rv = parse(in);
        skipWhitespace(in);
        if (!in.get(c) || c != ')')
            throw LLispException("read-syntax", "Expected ')'");
        return rv;
    } else {
        in.putback(c);
        LPtr cdr = parse(in);
        return new LCons(cdr, parseTail(in));
    }
}

LPtr
parse(std::istream& in)
{
    using namespace std;
    skipWhitespace(in);

    char c;
    if (!in.get(c))
        throw LLispException("eof", "End of file at beginning of expression");

    if (c == '(') {
        return parseTail(in);
    } else if (c == '\'') {
        return new LCons(new LSymbol("quote"), new LCons(parse(in), LPtr()));
    } else if (c == '"') {
        std::string accum;
        bool q = false;
        bool octal = 0;
        unsigned char caccum = 0;
        while (1) {
            if (!in.get(c))
                throw LLispException("read-syntax", "End of file in string constant");
            if (octal) {
                if (c >= '0' && c <= '7') {
                    caccum = 8*caccum + (c-'0');
                } else {
                    octal = false;
                    accum += char(caccum);
                }
            }
            if (octal) {
                /* nix */
            } else if (q) {
                if (c == 'r') {
                    accum += '\r';
                } else if (c == 'n') {
                    accum += '\n';
                } else if (c == 't') {
                    accum += '\t';
                } else if (c >= '0' && c <= '7') {
                    octal = true;
                    caccum = (c - '0');
                } else {
                    accum += c;
                }
                q = false;
            } else if (c == '"') {
                break;
            } else if (c == '\\') {
                q = true;
            } else {
                accum += c;
            }
        }
        return new LString(accum);
    } else {
        std::string accum;
        while (!(c == '(' || c == ')' || c == '\'' || c == '"' || c == ';' || c == '.' || isspace(c))) {
            accum += c;
            if (!in.get(c))
                goto have_it;
        }
        in.putback(c);
     have_it:
        if (!accum.size()) {
            in.get(c);
            throw LLispException("read-syntax", std::string("Parse error at '") + c + "'");
        }

        /* is it a number? special symbol? */
        int n;
        if (std::sscanf(accum.c_str(), "%d%c", &n, &c) == 1)
            return new LInt(n);
        else if (accum == "nil")
            return &sym_nil;
        else if (accum == "t")
            return &sym_t;
        else
            return new LSymbol(accum);
    }
}

/***************************** Special Forms *****************************/

//- (lambda (ARGS) BODY)
//- => Returns a function object taking (ARGS) with the given BODY.
//-    Example: (setq square (lambda (a) (* a a)))
//-    makes 'square' a function that squares its argument.
static LPtr
SFLambda(LPtr p)
{
    LCons* c = p.as<LCons>();
    if (!c || !isList(c->getCar()))
        throw LLispException("invalid-function", "Invalid function body in 'lambda'");
    return new LFunction(c->getCar(), c->getCdr());
}

//- (quote EXPRESSION)
//- => Returns the EXPRESSION, unevaluated. Can be abbreviated as 'EXPRESSION.
static LPtr
SFQuote(LPtr p)
{
    if (getListLength(p) != 1)
        throw LLispException("argument-count", "Too many arguments to 'quote'");
    return p.as<LCons>()->getCar();
}

//- (if COND THEN ELSE...)
//- => If COND evals to non-nil, evals THEN (and returns its value), otherwise,
//-    evals ELSE (and return the last expression's value). Whereas only one form
//-    is allowed for THEN, possibly many (or none) are allowed for ELSE.
static LPtr
SFIf(LPtr p)
{
    if (getListLength(p) < 2)
        throw LLispException("argument-count", "Too few arguments for 'if'");
    if (!getCar(p)->eval())
        return evalList(getTail(p, 2));
    else
        return getNth(p, 1)->eval();
}

//- (cond CLAUSE...)
//- => Multi-way condition. Each CLAUSE is a list of expressions. Each CLAUSE's
//-    first expression is evaluated until one yields non-nil. In this case, the
//-    remaining expressions of that clause, if any, are evaluated, and the value
//-    of the last one is returned. If no CLAUSE matches, returns nil.
//-    Example: (cond
//-                 ((eq x 1) (print "one"))
//-                 ((eq x 2) (print "two"))
//-                 (t        (print "neither one nor two")))
static LPtr
SFCond(LPtr p)
{
    LPtr result;
    bool hadOne = false;
    while (LCons* c = p.as<LCons>()) {
        /* iterate through all clauses */
        LPtr   elem = c->getCar();
        LCons* elemc = elem.as<LCons>();
        if (!elemc)
            throw LLispException("invalid-cond", "Malformed clause in 'cond'");

        if (!hadOne) {
            result = elemc->getCar()->eval();
            if (result.nonnull()) {
                /* found a clause */
                hadOne = true;
                if (elemc->getCdr().nonnull())
                    result = evalList(elemc->getCdr());
            }
        }

        p = c->getCdr();
    }

    if (p.nonnull())
        throw LLispException("malformed-list", "Malformed list");

    return result;
}

//- (while COND BODY....)
//- => Loop. While COND evals to non-nil, evaluate BODY. When COND yields nil,
//-    terminate the loop and return nil.
static LPtr
SFWhile(LPtr p)
{
    if (getListLength(p) < 1)
        throw LLispException("argument-count", "Too few arguments for 'while'");
    while (!!getCar(p)->eval())
        evalList(getCdr(p));
    return LPtr();
}

//- (foreach SYM LIST BODY...)
//- => Loop. Evaluate LIST, then iterate over all elements of list, binding
//-    SYM to the respective list element and eval BODY. Returns nil.
//-
//-    Note that a similar construct can also be implemented directly,
//-    so this one is mainly a convenience:
//-       (define (foreach* list fun)
//-         (while list
//-           (fun (car list))
//-           (setq list (cdr list)))
//-         )
//-       (foreach* '(1 2 3) (lambda (x) (print x)))
static LPtr
SFForeach(LPtr p)
{
    if (getListLength(p) < 2)
        throw LLispException("argument-count", "Too few arguments for 'foreach'");

    LPtr the_sym  = getCar(p);
    LPtr the_list = getNth(p, 1)->eval();
    p = getTail(p, 2);

    LSymbol* sym = the_sym.as<LSymbol>();
    if (!sym)
        throw LLispException("type-error", "First arg to 'foreach' must be symbol");

    Environment env;

    while (LCons* c = the_list.as<LCons>()) {
        env.set(sym->getSymbol(), c->getCar());
        env.bind();  
        evalList(p);
        the_list = c->getCdr();
    }
    return LPtr();
}

//- (or FORMS....)
//- => Eval FORMS until one returns non-nil. Then, return that value.
//-    If all forms eval to nil, return nil.
static LPtr
SFOr(LPtr p)
{
    while (p.nonnull()) {
        LPtr v = getCar(p)->eval();
        if (v.nonnull())
            return v;
        p = getCdr(p);
    }
    return LPtr();
}

//- (and FORMS....)
//- => Eval FORMS until one returns nil. Then, return nil. Otherwise,
//-    return last form's value.
static LPtr
SFAnd(LPtr p)
{
    LPtr v(&sym_t);
    while (p.nonnull()) {
        v = getCar(p)->eval();
        if (!v)
            return LPtr();
        p = getCdr(p);
    }
    return v;
}

//- (define SYMBOL VALUE)
//- (define (NAME ARGS) BODY)
//- => First form: evaluate VALUE, and assign it to SYMBOL. This form is
//-    equivalent to (set 'SYMBOL VALUE).
//-    Second form: the Scheme way of defining functions. Equivalent to
//-    (define NAME (lambda (ARGS) BODY)).
static LPtr
SFDefine(LPtr p)
{
    LCons* arg = p.as<LCons>();
    if (!arg)
        throw LLispException("argument-count", "'define' missing argument");

    if (LSymbol* sym = arg->getCar().as<LSymbol>()) {
        arg = arg->getCdr().as<LCons>();
        if (!arg)
            throw LLispException("argument-count", "'define' missing value");
        if (!!arg->getCdr())
            throw LLispException("argument-count", "Too many arguments to 'define'");
        LSymbol::define(sym->getSymbol(), arg->getCar()->eval());
    } else if (LCons* cons = arg->getCar().as<LCons>()) {
        LSymbol* name = cons->getCar().as<LSymbol>();
        if (!name)
            throw LLispException("argument-count", "'define' missing function name");
        if (!isList(cons->getCdr()))
            throw LLispException("malformed-list", "Invalid function body in 'define'");
        LSymbol::define(name->getSymbol(), new LFunction(cons->getCdr(), arg->getCdr()));
    } else {
        throw LLispException("argument-count", "Invalid form of 'define'");
    }
    return LPtr();
}

//- (progn FORMS...)
//- => Eval the FORMS, return the last form's value.
static LPtr
SFProgn(LPtr p)
{
    return evalList(p);
}

//- (prog1 FORMS...)
//- => Eval the forms, return the first form's value.
static LPtr
SFProg1(LPtr p)
{
    LCons* pp = p.as<LCons>();
    if (!pp)
        throw LLispException("malformed-list", "Argument to 'prog1' must be list");
    LPtr rv = pp->getCar()->eval();
    evalList(pp->getCdr());
    return rv;
}

//- (unwind-protect BODYFORMS UNWINDFORMS...)
//- => Evaluate BODYFORM, then UNWINDFORMS; returns BODYFORM's value.
//-    If BODYFORM exits non-locally (throw), evaluate the UNWINDFORMS anyway.
static LPtr
SFUnwindProtect(LPtr p)
{
    if (getListLength(p) < 2)
        throw LLispException("argument-count", "Too few arguments for 'unwind-protect'");
    try {
        LPtr rv = getCar(p)->eval();
        evalList(getCdr(p));
        return rv;
    }
    catch (...) {
        evalList(getCdr(p));
        throw;
    }
}

//- (catch SYM BODYFORMS...)
//- => Evaluate SYM, then the BODYFORMS; returns the last BODYFORM's value.
//-    If BODYFORMS exits by (throw SYM VALUE), return VALUE. If SYM is t,
//-    catch everything.
//-    Example: (catch 'foo (throw 'foo 1) 2)
//-    returns 1.
static LPtr
SFCatch(LPtr p)
{
    if (getListLength(p) < 2)
        throw LLispException("argument-count", "Too few arguments for 'catch'");
    LPtr sym1 = getCar(p)->eval();
    LSymbol* sym = sym1.as<LSymbol>();
    if (!sym && sym1 != &sym_t)
        throw LLispException("type-error", "Tag for 'catch' must be a symbol");
    try {
        return evalList(getCdr(p));
    }
    catch (LLispException& e) {
        if (sym == 0 || e.getTag() == sym->getSymbol())
            return e.getData();
        throw;
    }
}

static LPtr
doLet(LPtr p, bool seq)
{
    LPtr bindforms = getCar(p);
    if (!isList(bindforms))
        throw LLispException("type-error", "Invalid bindforms for 'let'");
    
    Environment env;
    while (bindforms.nonnull()) {
        LPtr form = getCar(bindforms);
        if (LCons* c = form.as<LCons>()) {
            LSymbol* s = c->getCar().as<LSymbol>();
            LCons* c2 = c->getCdr().as<LCons>();
            if (!s || !c2 || c2->getCdr().nonnull())
                throw LLispException("type-error", "Invalid bindform in 'let'");
            env.set(s->getSymbol(), c2->getCar()->eval());
        } else if (LSymbol* s = form.as<LSymbol>()) {
            env.set(s->getSymbol(), LPtr());
        } else {
            throw LLispException("type-error", "Invalid bindform in 'let'");
        }
        if (seq)
            env.bind();
        bindforms = getCdr(bindforms);
    }
    if (!seq)
        env.bind();
    return evalList(getCdr(p));
}

//- (let (BINDFORMS...) BODYFORMS...)
//- => Process all BINDFORMS and change the variables accordingly; then
//-    evaluate the BODYFORMS with the changed bindings. Before let exits
//-    (normally or nonlocally), restores the previous values.
//-    Each BINDFORM is either a SYMBOL (which is bound to nil by let),
//-    or (SYMBOL EXPR) (EXPR is evaluated and bound to SYMBOL).
//-    All SYMBOLs are bound simultaneously, after evaluating all the EXPRs.
static LPtr
SFLet(LPtr p)
{
    return doLet(p, false);
}

//- (let* (BINDFORMS...) BODYFORMS...)
//- => Process all BINDFORMS and change the variables accordingly; then
//-    evaluate the BODYFORMS with the changed bindings. Before let exits
//-    (normally or nonlocally), restores the previous values.
//-    Each BINDFORM is either a SYMBOL (which is bound to nil by let),
//-    or (SYMBOL EXPR) (EXPR is evaluated and bound to SYMBOL).
//-    SYMBOLs are bound in the order they appear, so later EXPRs can refer
//-    to previous SYMBOLs.
static LPtr
SFLetS(LPtr p)
{
    return doLet(p, true);
}

//- (setf (EXPR) VALUE ...)
//- => Make (EXPR) return VALUE. Works only if an appropriate definition
//-    has been made; see 'define-setf'.
//-    For example, '(setf (car x) y)' is the same as '(setcar x y)'.
static LPtr
SFSetf(LPtr p)
{
    do {
        LCons* c = p.as<LCons>();
        if (!c)
            throw LLispException("argument-count", "Too few arguments to 'setf'");
        LPtr ex = c->getCar();
        c = c->getCdr().as<LCons>();
        if (!c)
            throw LLispException("argument-count", "Too few arguments to 'setf'");
        ex->setf(c->getCar());
        p = c->getCdr();
    }
    while (p.nonnull());
    return LPtr();
}

//- (setq SYM VALUE ...)
//- => Set SYM to VALUE. Same as (set 'sym value). Returns last value.
//-    Assignments happen in order, not simultaneously.
static LPtr
SFSetq(LPtr p)
{
    LPtr rv;
    do {
        LCons* c = p.as<LCons>();
        if (!c)
            throw LLispException("argument-count", "Too few arguments to 'setq'");
        LPtr ex = c->getCar();
        LSymbol* sym = ex.as<LSymbol>();
        if (!sym)
            throw LLispException("type-error", "Argument to 'setq' must be symbol");
        c = c->getCdr().as<LCons>();
        if (!c)
            throw LLispException("argument-count", "Too few arguments to 'setq'");
        rv = c->getCar()->eval();
        LSymbol::define(sym->getSymbol(), rv);
        p = c->getCdr();
    }
    while (p.nonnull());
    return rv;
}

/******************************** By-Value *******************************/

LPtr
getArg(LPtr& p)
{
    LCons* c = p.as<LCons>();
    if (!c)
        throw LLispException("argument-count", "Too few arguments");
    p = c->getCdr();
    return c->getCar();
}

int
argInt(LPtr& p)
{
    LPtr iv = getArg(p);
    LInt* i = iv.as<LInt>();
    if (!i)
        throw LLispException("type-error", "Expected integer");
    return i->getVal();
}

std::string
argString(LPtr& p)
{
    LPtr sv = getArg(p);
    LString* s = sv.as<LString>();
    if (!s)
        throw LLispException("type-error", "Expected string");
    return s->getVal();
}

std::string
argSym(LPtr& p)
{
    LPtr sv = getArg(p);
    LSymbol* s = sv.as<LSymbol>();
    if (!s)
        throw LLispException("type-error", "Expected symbol");
    return s->getSymbol();
}

bool
argBool(LPtr& p)
{
    LPtr bv = getArg(p);
    return bv != &sym_nil;
}

LPtr
concatRest(LPtr p)
{
    LCons* c = p.as<LCons>();
    if (!c)
        return LPtr();
    else if (!c->getCdr())
        return c->getCar();
    else
        return new LCons(c->getCar(), concatRest(c->getCdr()));
}

//- (+ N...)
//- => Add all parameters (integers).
static LPtr
FPlus(LPtr p)
{
    int n = 0;
    while (p.nonnull()) {
        n += argInt(p);
    }
    return new LInt(n);
}

//- (logior N...)
//- => Yields bitwise OR of all parameters (integers).
static LPtr
FLogIOR(LPtr p)
{
    int n = 0;
    while (p.nonnull()) {
        n |= argInt(p);
    }
    return new LInt(n);
}

//- (logxor N...)
//- => Yields bitwise XOR of all parameters (integers).
static LPtr
FLogXOR(LPtr p)
{
    int n = 0;
    while (p.nonnull()) {
        n ^= argInt(p);
    }
    return new LInt(n);
}

//- (- A B....)
//- => Subtract all B's from A (integers).
static LPtr
FMinus(LPtr p)
{
    int n = argInt(p);
    while (p.nonnull()) {
        n -= argInt(p);
    }
    return new LInt(n);
}

//- (logand N...)
//- => Yields bitwise AND of all parameters (integers).
static LPtr
FLogAND(LPtr p)
{
    unsigned int n = -1U;
    while (p.nonnull()) {
        n &= argInt(p);
    }
    return new LInt(n);
}

//- (lognot N)
//- => Yields bitwise negation of N.
static LPtr
FLogNOT(LPtr p)
{
    return new LInt(~argInt(p));
}

//- (* N...)
//- => Multiply all parameters.
static LPtr
FMult(LPtr p)
{
    int n = 1;
    while (p.nonnull()) {
        n *= argInt(p);
    }
    return new LInt(n);
}

//- (/ A B....)
//- => Divide A by all B's.
static LPtr
FDiv(LPtr p)
{
    int n = argInt(p);
    while (p.nonnull()) {
        int x = argInt(p);
        if (!x)
            throw LLispException("divide-error", "Division by zero");
        n /= x;
    }
    return new LInt(n);
}

//- (% A B....)
//- => Return remainder of A divided by all B's.
static LPtr
FMod(LPtr p)
{
    int n = argInt(p);
    while (p.nonnull()) {
        int x = argInt(p);
        if (!x)
            throw LLispException("divide-error", "Division by zero");
        n %= x;
    }
    return new LInt(n);
}

//- (< A B)
//- => Returns t if A < B, otherwise nil.
static LPtr
FLt(LPtr p)
{
    int a = argInt(p), b = argInt(p);
    if (a < b)
        return LPtr(&sym_t);
    else
        return LPtr();
}

//- (> A B)
//- => Returns t if A > B, otherwise nil.
static LPtr
FGt(LPtr p)
{
    int a = argInt(p), b = argInt(p);
    if (a > b)
        return LPtr(&sym_t);
    else
        return LPtr();
}

//- (<= A B)
//- => Returns t if A <= B, otherwise nil.
static LPtr
FLe(LPtr p)
{
    int a = argInt(p), b = argInt(p);
    if (a <= b)
        return LPtr(&sym_t);
    else
        return LPtr();
}

//- (>= A B)
//- => Returns t if A >= B, otherwise nil.
static LPtr
FGe(LPtr p)
{
    int a = argInt(p), b = argInt(p);
    if (a >= b)
        return LPtr(&sym_t);
    else
        return LPtr();
}

//- (set SYM VALUE SYM VALUE...)
//- => Set SYM to VALUE, e.g. (set 'foo 1). Returns last VALUE.
static LPtr
FSet(LPtr p)
{
    LPtr rv;
    while (p.nonnull()) {
        std::string name = argSym(p);
        rv = getArg(p);
        LSymbol::define(name, rv);
    }
    return rv;
}

//- (not VAL)
//- => If VAL evals to nil, return t, otherwise return nil.
static LPtr
FNot(LPtr p)
{
    if (argBool(p))
        return LPtr();
    else
        return LPtr(&sym_t);
}

//- (cons A B)
//- => Build new cons cell out of A and B.
static LPtr
FCons(LPtr p)
{
    LPtr a = getCar(p);
    LPtr b = getNth(p,1);
    return new LCons(a, b);
}

//- (car A)
//- => Returns the car of A, i.e. (car (cons X Y)) yields X.
static LPtr
FCar(LPtr p)
{
    return getCar(getCar(p));
}

//- (cdr A)
//- => Returns the cdr of A, i.e. (cdr (cons X Y)) yields Y.
static LPtr
FCdr(LPtr p)
{
    return getCdr(getCar(p));
}

//- (setcar C VAL)
//- => Modify C's car in place.
static LPtr
FSetCar(LPtr p)
{
    LPtr a = getArg(p), b = getArg(p);
    if (LCons* c = a.as<LCons>()) {
        c->setCar(b);
        return a;
    } else {
        throw LLispException("not-cons", "Not cons");
    }
}

//- (setcdr C VAL)
//- => Modify C's cdr in place
static LPtr
FSetCdr(LPtr p)
{
    LPtr a = getArg(p), b = getArg(p);
    if (LCons* c = a.as<LCons>()) {
        c->setCdr(b);
        return a;
    } else {
        throw LLispException("not-cons", "Not cons");
    }
}

//- (print VALUE)
//- => Print VALUE readably, on stdout, followed by a newline. This
//-    can be used for debugging.
static LPtr
FPrint(LPtr p)
{
    LPtr rv = getCar(p);
    std::cout << rv->toString() << "\n";
    return rv;
}

//- (print-string STRING...)
//- => Print all parameters which must be strings.
static LPtr
FPrintString(LPtr p)
{
    while (p.nonnull()) {
        std::cout << argString(p);
    }
    return LPtr();
}

//- (read [str])
//- => Read an expression from stdin and return that. If an argument is
//-    given, it must be a string; read will read that string.
static LPtr
FRead(LPtr p)
{
    if (p.nonnull()) {
        std::stringstream in(argString(p));
        return parse(in);
    } else {
        return parse(std::cin);
    }
}

//- (load STRING)
//- => Load and evaluate the file given by STRING.
static LPtr
FLoad(LPtr p)
{
    std::string s = argString(p);
    std::ifstream in(s.c_str());
    if (!in)
        throw LLispException("file-open", "Cannot open file");
    LPtr rv;
    while (1) {
        LPtr tmp;
        try {
            tmp = parse(in);
        }
        catch (LLispException& e) {
            if (e.getTag() == "eof")
                break;
            throw;
        }
        rv = tmp->eval();
    }
    return rv;
}

//- (apply FUNCTION ARG ARG ARG... REST)
//- => Builds an argument list out of the ARGS. The last argument, REST, must
//-    be a list. Then, applies FUNCTION to these arguments.
//-    Example: (apply cons 1 2 nil)
//-    Same as: (apply cons '(1 2))
//-    Same as: (cons 1 2)
//-    Note that if the argument list is fixed, you can even place the function
//-    first in the list, as in (progn (setq f cons) (f 1 2)), you need not use
//-    apply to simulate function variables. However, this is useful to call a
//-    function with variable arguments, e.g. (apply concat LIST-OF-STRINGS).
static LPtr
FApply(LPtr p)
{
    LPtr fun = getArg(p);
    LPtr rest = concatRest(p);
    LApplyByValue* abv = fun.as<LApplyByValue>();
    if (!abv)
        throw LLispException("type-error", "Invalid function in 'apply'");
    return abv->applyEvaluated(rest);
}

//- (equal A B)
//- => t iff A and B are structurally identical, for example, two lists of
//-    the same contents.
static LPtr
FEqual(LPtr p)
{
    LPtr a = getArg(p);
    LPtr b = getArg(p);
    if (a->equals(b, false))
        return LPtr(&sym_t);
    else
        return LPtr();
}

//- (/= A B)
//- => t iff A and B are not structurally identical. This is the opposite of
//-    (equal A B).
static LPtr
FNe(LPtr p)
{
    LPtr a = getArg(p);
    LPtr b = getArg(p);
    if (!a->equals(b, false))
        return LPtr(&sym_t);
    else
        return LPtr();
}

//- (eq A B)
//- => t iff A and B are identical, i.e. the same object. This tries to
//-    simulate the Emacs Lisp behaviour. Two strings with the same
//-    contents do not necessarily compare equal if they are different
//-    objects. Two symbols of the same name are guaranteed to compare
//-    equal.
static LPtr
FEq(LPtr p)
{
    LPtr a = getArg(p);
    LPtr b = getArg(p);
    if (a->equals(b, true))
        return LPtr(&sym_t);
    else
        return LPtr();
}

//- (concat A...)
//- => All parameters are strings. Concatenates them and returns the result.
static LPtr
FConcat(LPtr p)
{
    std::string s;
    while (p.nonnull()) {
        s += argString(p);
    }
    return new LString(s);
}

//- (substring S START [LEN])
//- => Extract substring of S. Indexing starts at 0. If LEN is omitted,
//-    yields the complete rest of S starting at START.
static LPtr
FSubstring(LPtr p)
{
    // FIXME: (substring a b nil) should be identical to (substring a b)
    std::string s                = argString(p);
    std::string::size_type start = argInt(p);
    std::string::size_type len   = p.nonnull() ? argInt(p) : std::string::npos;
    if (start > s.size())
        throw LLispException("range-error", "String index out of range");
    return new LString(s.substr(start, len));
}

//- (string= a b)
//- => t iff A equals B. Same as (equal), but makes sure the parameters are strings.
static LPtr
FStringEq(LPtr p)
{
    std::string a = argString(p);
    std::string b = argString(p);
    if (a == b)
        return LPtr(&sym_t);
    else
        return LPtr();
}

//- (string< a b)
//- => t iff A lexicographically precedes B.
static LPtr
FStringLt(LPtr p)
{
    std::string a = argString(p);
    std::string b = argString(p);
    if (a < b)
        return LPtr(&sym_t);
    else
        return LPtr();
}

//- (string-find HAYSTACK NEEDLE [START])
//- => return first position where NEEDLE appears in HAYSTACK, indexing starts at
//-    zero. If START is given, return no less than START. Returns the string
//-    position, or nil if no such position.
static LPtr
FStringFind(LPtr p)
{
    std::string a = argString(p);
    std::string n = argString(p);
    std::string::size_type start = p.nonnull() ? argInt(p) : 0;
    if (start > a.size())
        throw LLispException("range-error", "String index out of range");
    std::string::size_type found = a.find(n, start);
    if (found != a.npos)
        return new LInt(found);
    else
        return LPtr();
}

//- (string-to-upper S)
//- => convert string to upper-case.
static LPtr
FStringToUpper(LPtr p)
{
    using std::toupper;
    std::string a = argString(p);
    for (std::string::size_type i = 0; i < a.size(); ++i)
        a[i] = toupper((unsigned char) a[i]);
    return new LString(a);
}

//- (string-to-lower S)
//- => convert string to lower-case.
static LPtr
FStringToLower(LPtr p)
{
    using std::tolower;
    std::string a = argString(p);
    for (std::string::size_type i = 0; i < a.size(); ++i)
        a[i] = tolower((unsigned char) a[i]);
    return new LString(a);
}

//- (string-length S)
//- => return length of S.
static LPtr
FStringLength(LPtr p)
{
    return new LInt(argString(p).size());
}

//- (print-to-string S)
//- => Convert the parameter into its readable string representation.
//-    Compare (print).
static LPtr
FPrintToString(LPtr p)
{
    return new LString(getCar(p)->toString());
}

//- (bye [exitcode])
//- => Exit the interpreter, with no cleanup (i.e. unwind-protected code
//-    is not executed). Returns the specified exitcode, 0 if none given.
static LPtr
FBye(LPtr p)
{
    int result = 0;
    if (p.nonnull())
        result = argInt(p);
    std::exit(result);
    return LPtr();
}

//- (throw SYM DATA)
//- => Throws an exception that can be caught by (catch SYM ...). Note
//-    that SYM is evaluated, so either specify a no-eval symbol, or
//-    quote it.
static LPtr
FThrow(LPtr p)
{
    std::string tag = argSym(p);
    LPtr data = getArg(p);
    throw LLispException(tag, data);
}

//- (stringp x), (consp x), (integerp x), (symbolp x), (functionp x), (builtinp x)
//- => Return t if x is of the correct type, nil otherwise.
//-    functionp returns t for built-in and user-defined functions,
//-    builtinp returns t for built-in functions only. Neither returns
//-    t for special forms.
template<typename N>
static LPtr
FTestp(LPtr p)
{
    if (getArg(p).as<N>())
        return LPtr(&sym_t);
    else
        return LPtr();
}

//- (gensym)
//- => Return a new symbol that is currently not bound. Since we
//-    do not intern symbols like real Lisp does, this does not
//-    guarantee to avoid clashes with symbols used later on. However,
//-    this implementation of gensym is guaranteed to never return a
//-    symbol twice even if it is not immediately bound, so you can
//-    pre-allocate symbols.
static LPtr
FGensym(LPtr /*p*/)
{
    static int counter = 0;
    char tmp[20];
    do {
        std::sprintf(tmp, "G%06d", ++counter);
    } while (symtab.find(std::string(tmp)) != symtab.end());
    return new LSymbol(tmp);
}

//- (eval X)
//- => Evaluate expression X. For example, (eval '(+ 2 3)) returns 5.
static LPtr
FEval(LPtr p)
{
    return getCar(p)->eval();
}

//- (boundp X)
//- => t iff symbol X is bound, i.e. has been set using set, setq
//-    or define.
static LPtr
FBoundp(LPtr p)
{
    std::string s = argSym(p);
    if (symtab.find(s) != symtab.end())
        return LPtr(&sym_t);
    else
        return LPtr();
}

//- (list ELEMS...)
//- => return list made of all ELEMS. For example,
//-    (list 1 2 3) is the same as '(1 2 3).
static LPtr
FList(LPtr p)
{
    return p;
}

//- (define-setf GETTER SETTER)
//- => Arranges for (setf (GETTER ARGS) Y) to call (SETTER ARGS Y).
//-    Both getter and setter must be functions, not special forms.
static LPtr
FDefineSetf(LPtr p)
{
    LPtr s = getArg(p);
    LPtr f = getArg(p);

    LApplyable* a = s.as<LApplyable>();
    if (!a)
        throw LLispException("invalid-setf", "Not a function");
    a->defineSet(f);
    return f;
}

//- (mapatoms FUNC)
//- => Call (FUNC X) for each symbol X that is currently bound.
//-    Example: (mapatoms print)
//-    prints the names of all symbols.
static LPtr
FMapatoms(LPtr p)
{
    LPtr s = getArg(p);
    LApplyByValue* a = s.as<LApplyByValue>();
    if (!a)
        throw LLispException("invalid-function", "Cannot apply this in 'mapatoms'");
    for (std::map<std::string,LPtr>::iterator i = symtab.begin(); i != symtab.end(); ++i)
        a->applyEvaluated(new LCons(new LSymbol(i->first), LPtr()));
    return LPtr();
}

//- (mapstring FUNC STRING)
//- => Call (FUNC X) for each character of the STRING. Concatenates the
//-    results, which must all be strings or nil.
//-    Example: (mapstring (lambda (x) (concat x "_")) "foo")
//-    returns "f_o_o_".
static LPtr
FMapstring(LPtr p)
{
    LPtr s = getArg(p);
    std::string arg = argString(p);
    std::string result;

    LApplyByValue* a = s.as<LApplyByValue>();
    if (!a)
        throw LLispException("invalid-function", "Cannot apply this in 'mapstring'");
    for (std::string::size_type i = 0; i < arg.size(); ++i) {
        LPtr val = a->applyEvaluated(new LCons(new LString(std::string(arg, i, 1)), LPtr()));
        if (!val) {
            // empty
        } else if (LString* ps = val.as<LString>()) {
            result += ps->getVal();
        } else {
            throw LLispException("type-error", "Need string result in 'mapstring'");
        }
    }
    return new LString(result);
}


//- (int-to-char NUMBER)
//- => Return the character (i.e. one-character string) corresponding
//-    to ASCII code NUMBER
static LPtr
FIntToChar(LPtr p)
{
    int n = argInt(p);
    return new LString(std::string(1, char(n)));
}

//- (char-to-int STRING)
//- => Return the ASCII code of the first character of STRING,
//-    or nil if STRING is empty
static LPtr
FCharToInt(LPtr p)
{
    std::string s = argString(p);
    if (s.size())
        return new LInt(static_cast<unsigned char>(s[0]));
    else
        return LPtr();
}

//- (string-to-char-list STRING)
//- => Splits STRING into single characters and returns list of these.
static LPtr
FStringToCharList(LPtr p)
{
    std::string s = argString(p);
    LPtr result;
    for (std::string::size_type i = s.size(); i > 0; --i)
        result = new LCons(new LString(std::string(1, s[i-1])), result);
    return result;
}

//- (rand N)
//- => Generate random number from [0,N)
static LPtr
FRand(LPtr p)
{
    int n = argInt(p);
#if RAND_MAX < 65536
    return new LInt(((unsigned long) std::rand() * (unsigned long) n) / RAND_MAX);
#else
    return new LInt(int((std::rand() * ((double) n / RAND_MAX))));
#endif
}

//- (srand [N])
//- => Seed random number generator to N (or current time).
static LPtr
FSRand(LPtr p)
{
    if (p.nonnull())
        std::srand(argInt(p));
    else
        std::srand(std::time(0));
    return LPtr();
}

//- (function-definition FUNC)
//- => Returns function FUNC's definition as a list.
//-    Example: (function-definition (lambda (x) (foo) (bar)) = ((foo) (bar))
static LPtr
FFunctionDefinition(LPtr p)
{
    LPtr a = getArg(p);
    if (LFunction* func = a.as<LFunction>()) {
        return func->getBody();
    } else {
        return LPtr();
    }
}

//- (function-arglist FUNC)
//- => Returns function FUNC's argument list
//-    Example: (function-arglist (lambda (x) (foo) (bar)) = (x)
static LPtr
FFunctionArglist(LPtr p)
{
    LPtr a = getArg(p);
    if (LFunction* func = a.as<LFunction>()) {
        return func->getArgs();
    } else {
        return LPtr();
    }
}

//- newline
//- => A string containing a lone newline character (this precedes the
//-    introduction of "\r" sequences).

/********************************** main *********************************/

void
initBasicSymbols()
{
    // values
    LSymbol::define("nil",             &sym_nil);
    LSymbol::define("t",               &sym_t);
    LSymbol::define("newline",         new LString("\n"));

    // special forms
    LSymbol::define("lambda",          new LSpecial(SFLambda));
    LSymbol::define("quote",           new LSpecial(SFQuote));
    LSymbol::define("if",              new LSpecial(SFIf));
    LSymbol::define("cond",            new LSpecial(SFCond));
    LSymbol::define("while",           new LSpecial(SFWhile));
    LSymbol::define("foreach",         new LSpecial(SFForeach));
    LSymbol::define("and",             new LSpecial(SFAnd));
    LSymbol::define("or",              new LSpecial(SFOr));
    LSymbol::define("define",          new LSpecial(SFDefine));
    LSymbol::define("progn",           new LSpecial(SFProgn));
    LSymbol::define("prog1",           new LSpecial(SFProg1));
    LSymbol::define("unwind-protect",  new LSpecial(SFUnwindProtect));
    LSymbol::define("catch",           new LSpecial(SFCatch));
    LSymbol::define("let",             new LSpecial(SFLet));
    LSymbol::define("let*",            new LSpecial(SFLetS));
    LSymbol::define("setf",            new LSpecial(SFSetf));
    LSymbol::define("setq",            new LSpecial(SFSetq));

    // functions
    LSymbol::define("+",               new LBuiltin(FPlus,             0, INT_MAX));
    LSymbol::define("logior",          new LBuiltin(FLogIOR,           0, INT_MAX));
    LSymbol::define("logxor",          new LBuiltin(FLogXOR,           0, INT_MAX));
    LSymbol::define("-",               new LBuiltin(FMinus,            1, INT_MAX));
    LSymbol::define("logand",          new LBuiltin(FLogAND,           0, INT_MAX));
    LSymbol::define("lognot",          new LBuiltin(FLogNOT,           1, 1));
    LSymbol::define("*",               new LBuiltin(FMult,             0, INT_MAX));
    LSymbol::define("/",               new LBuiltin(FDiv,              1, INT_MAX));
    LSymbol::define("%",               new LBuiltin(FMod,              1, INT_MAX));
    LSymbol::define("<",               new LBuiltin(FLt,               2, 2));
    LSymbol::define(">",               new LBuiltin(FGt,               2, 2));
    LSymbol::define("<=",              new LBuiltin(FLe,               2, 2));
    LSymbol::define(">=",              new LBuiltin(FGe,               2, 2));
    LSymbol::define("/=",              new LBuiltin(FNe,               2, 2));
    LSymbol::define("set",             new LBuiltin(FSet,              2, INT_MAX));
    LSymbol::define("not",             new LBuiltin(FNot,              1, 1));
    LSymbol::define("cons",            new LBuiltin(FCons,             2, 2));
    LSymbol::define("car",             new LBuiltin(FCar,              1, 1));
    LSymbol::define("cdr",             new LBuiltin(FCdr,              1, 1));
    LSymbol::define("setcar",          new LBuiltin(FSetCar,           2, 2));
    LSymbol::define("setcdr",          new LBuiltin(FSetCdr,           2, 2));
    LSymbol::define("print",           new LBuiltin(FPrint,            1, 1));
    LSymbol::define("print-string",    new LBuiltin(FPrintString,      0, INT_MAX));
    LSymbol::define("read",            new LBuiltin(FRead,             0, 1));
    LSymbol::define("load",            new LBuiltin(FLoad,             1, 1));
    LSymbol::define("apply",           new LBuiltin(FApply,            1, INT_MAX));
    LSymbol::define("eq",              new LBuiltin(FEq,               2, 2));
    LSymbol::define("equal",           new LBuiltin(FEqual,            2, 2));
    LSymbol::define("concat",          new LBuiltin(FConcat,           0, INT_MAX));
    LSymbol::define("substring",       new LBuiltin(FSubstring,        2, 3));
    LSymbol::define("string=",         new LBuiltin(FStringEq,         2, 2));
    LSymbol::define("string<",         new LBuiltin(FStringLt,         2, 2));
    LSymbol::define("string-find",     new LBuiltin(FStringFind,       2, 3));
    LSymbol::define("string-to-upper", new LBuiltin(FStringToUpper,    1, 1));
    LSymbol::define("string-to-lower", new LBuiltin(FStringToLower,    1, 1));
    LSymbol::define("string-length",   new LBuiltin(FStringLength,     1, 1));
    LSymbol::define("print-to-string", new LBuiltin(FPrintToString,    1, 1));
    LSymbol::define("bye",             new LBuiltin(FBye,              0, 1));
    LSymbol::define("throw",           new LBuiltin(FThrow,            2, 2));
    // Borland C++ does not accept this without the explicit construction
    LSymbol::define(std::string("consp"),     new LBuiltin(FTestp<LCons>,     1, 1));
    LSymbol::define(std::string("stringp"),   new LBuiltin(FTestp<LString>,   1, 1));
    LSymbol::define(std::string("integerp"),  new LBuiltin(FTestp<LInt>,      1, 1));
    LSymbol::define(std::string("symbolp"),   new LBuiltin(FTestp<LSymbol>,   1, 1));
    LSymbol::define(std::string("builtinp"),  new LBuiltin(FTestp<LBuiltin>,  1, 1));
    LSymbol::define(std::string("functionp"), new LBuiltin(FTestp<LApplyByValue>, 1, 1));
    LSymbol::define("gensym",          new LBuiltin(FGensym,           0, 0));
    LSymbol::define("eval",            new LBuiltin(FEval,             1, 1));
    LSymbol::define("boundp",          new LBuiltin(FBoundp,           1, 1));
    LSymbol::define("list",            new LBuiltin(FList,             0, INT_MAX));
    LSymbol::define("define-setf",     new LBuiltin(FDefineSetf,       2, 2));
    LSymbol::define("mapatoms",        new LBuiltin(FMapatoms,         1, 1));
    LSymbol::define("mapstring",       new LBuiltin(FMapstring,        2, 2));
    LSymbol::define("int-to-char",     new LBuiltin(FIntToChar,        1, 1));
    LSymbol::define("char-to-int",     new LBuiltin(FCharToInt,        1, 1));
    LSymbol::define("string-to-char-list", new LBuiltin(FStringToCharList, 1, 1));
    LSymbol::define("rand",            new LBuiltin(FRand,             1, 1));
    LSymbol::define("srand",           new LBuiltin(FSRand,            0, 1));
    LSymbol::define("function-definition", new LBuiltin(FFunctionDefinition, 1, 1));
    LSymbol::define("function-arglist", new LBuiltin(FFunctionArglist, 1, 1));

    LSymbol::defineSetf("car", "setcar");
    LSymbol::defineSetf("cdr", "setcdr");
}

#ifdef MAIN
int
main(int /*argc*/, char** /*argv*/)
{
    initBasicSymbols();
    while (1) {
        try {
            LPtr x = parse(std::cin);
            std::cout << "=> " << x->eval()->toString() << std::endl;
        }
        catch (LLispException& e) {
            if (e.getTag() == "eof")
                break;

            std::cerr << "error: " << e.getData()->toString() << std::endl;
        }
    }
    return 0;
}
#endif
