/**
  *  \file lispy.h
  *  \brief Not Quite Lisp Nor Scheme
  *
  *  (c) 2005 Stefan Reuther <Streu@gmx.de>
  */
#ifndef LISPY_H_INCLUDED
#define LISPY_H_INCLUDED

#include <string>
#include <istream>

class LPtr;

/// Basic Language Object.
class LObject {
    friend class LPtr;
    friend class LSelfEvalSym;
    int ref;
    LObject(const LObject&);
    LObject& operator=(const LObject&);
 public:
    LObject()
        : ref(0)
        { }
    virtual ~LObject();
    virtual LPtr eval() = 0;
    virtual std::string toString() = 0;
    virtual bool equals(LPtr other, bool identical);
    virtual void setf(LPtr value) = 0;
};

/// Self-evaluating Object.
class LSelfEval : public LObject {
 public:
    ~LSelfEval();
    LPtr eval();
    void setf(LPtr value);
};

/// Self-evaluating Symbol.
class LSelfEvalSym : public LSelfEval {
    const char* name;
 public:
    LSelfEvalSym(const char* name);
    std::string toString();
};

extern LSelfEvalSym sym_nil, sym_t;

/// Internal reference to a LObject.
class LPtr {
    LObject* p;
 public:
    LPtr();
    LPtr(const LPtr&);
    LPtr(LObject*);
    ~LPtr();

    LPtr& operator=(const LPtr&);

    bool operator==(const LPtr& rhs) const
        { return p == rhs.p; }
    bool operator!=(const LPtr& rhs) const
        { return p != rhs.p; }
    bool operator!() const
        { return p == &sym_nil; }
    bool nonnull() const
        { return p != &sym_nil; }

    LObject* operator->() const
        { return p; }

    template<typename T>
    T*
    as() const
        { return dynamic_cast<T*>(p); }
};

/// Integer object.
class LInt : public LSelfEval {
    int n;
 public:
    LInt(int n)
        : n(n)
        { }
    ~LInt();
    std::string toString();
    int getVal() const
        { return n; }
    bool equals(LPtr other, bool identical);
};

/// String object.
class LString : public LSelfEval {
    std::string s;
 public:
    LString(std::string s)
        : s(s)
        { }
    ~LString();
    std::string toString();
    std::string getVal() const
        { return s; }
    bool equals(LPtr other, bool identical);
};

/// Symbol.
class LSymbol : public LObject {
    std::string s;
 public:
    LSymbol(std::string s)
        : s(s)
        { }
    ~LSymbol();
    LPtr eval();
    std::string toString();
    std::string getSymbol() const
        { return s; }
    bool equals(LPtr other, bool identical);
    void setf(LPtr p);

    static void define(std::string name, LPtr value);
    static void defineSetf(std::string get, std::string set);
};

/// Applyable object. Something that can be used as a function.
class LApplyable : public LObject {
    std::string name;
    LPtr setf_function;
 public:
    virtual LPtr apply(LPtr args) = 0;
    void applySet(LPtr args, LPtr val);
    void setNameMaybe(std::string name);
    std::string getName() const;
    void defineSet(LPtr args);
    void setf(LPtr val);
};

/// List node.
class LCons : public LObject {
    LPtr car, cdr;
 public:
    LCons(LPtr car, LPtr cdr)
        : car(car), cdr(cdr)
        { }
    ~LCons();
    LPtr eval();
    std::string toString();
    LPtr getCar() const
        { return car; }
    LPtr getCdr() const
        { return cdr; }
    void setCdr(LPtr p)
        { cdr = p; }
    void setCar(LPtr p)
        { car = p; }
    bool equals(LPtr other, bool identical);
    void setf(LPtr value);
};

/// Special form / builtin function.
class LSpecial : public LApplyable {
    LPtr (*fun)(LPtr);
 public:
    LSpecial(LPtr(*fun)(LPtr))
        : fun(fun)
        { }
    ~LSpecial();
    LPtr eval()
        { return this; }
    LPtr apply(LPtr p);
    std::string toString();
};

/// Function taking by-value parameters.
class LApplyByValue : public LApplyable {
 public:
    LPtr eval()
        { return this; }
    LPtr apply(LPtr p);
    virtual LPtr applyEvaluated(LPtr p) = 0;
};

/// Built-in function.
class LBuiltin : public LApplyByValue {
    LPtr (*fun)(LPtr);
    int min, max;
 public:
    LBuiltin(LPtr(*fun)(LPtr), int min, int max)
        : fun(fun), min(min), max(max)
        { }
    ~LBuiltin();
    LPtr applyEvaluated(LPtr p);
    std::string toString();
};

/// Function object. Result of a lambda expression.
class LFunction : public LApplyByValue {
    LPtr args, body;
 public:
    LFunction(LPtr args, LPtr body);
    ~LFunction();
    LPtr applyEvaluated(LPtr p);
    std::string toString()
        { return "#<function:" + getName() + ">"; }
    LPtr getBody()
        { return body; }
    LPtr getArgs()
        { return args; }
};

class LLispException : public std::exception {
    std::string tag;
    LPtr        data;
    LPtr        trace;
 public:
    LLispException(std::string tag, std::string message)
        : tag(tag), data(new LString(message))
        { }
    LLispException(std::string tag, LPtr data)
        : tag(tag), data(data)
        { }
    ~LLispException() throw()
        { }
    const char* what() const throw()
        { return tag.c_str(); }
    std::string getTag() const
        { return tag; }
    LPtr getData() const
        { return data; }
    void pushTrace(LPtr obj);
    LPtr getTrace()
        { return trace; }
};

LPtr parse(std::istream& in);

void initBasicSymbols();

LPtr getArg(LPtr& p);
int argInt(LPtr& p);
std::string argString(LPtr& p);
std::string argSym(LPtr& p);
bool argBool(LPtr& p);

#endif
