#!/usr/bin/perl -w
#
#  This reads a text file and drives the lisp interpreter with the
#  expressions contained there-in, expecting a the specified answers.
#
#  Example:
#     (+ 1 2)
#     > => 3
#  means to feed '(+ 1 2)' into the interpreter and expect '=> 3' as
#  an answer. The expression can be multiple lines, as can the answer
#  (each line must be introduced with a '>').
#
#  Caveat: this starts just one instance of the interpreter and feeds
#  it input and output. When the expected answer is N lines, this
#  will read N lines from the interpreter, and hang if the actual
#  answer is shorter.
#
#  A possible workaround would be to start a new interpreter for every
#  test case. This would make this tester more reliable to errors, and
#  even allow it to run on DOS (by using files for communication), but
#  would also require each individual test to be completely
#  self-contained.
#

use strict;
use FileHandle;
use IPC::Open2;

$| = 1;

my $pid = open2(*Reader, *Writer, "./lispy 2>&1");

my $state = 'pre';
my $command = "";
my $firstLine;
my @result;
my $numTests = 0;
while (<>) {
    s/^ *;.*//;
    if ($state eq 'pre') {
        if (/^\s*$/) {
            # nix
        } else {
            $command = $_;
            $state = 'command';
            $firstLine = $.;
        }
    } elsif ($state eq 'command') {
        if (/^> *(.*)/) {
            $state = 'result';
            @result = "$1\n";
        } else {
            $command .= $_;
        }
    } elsif ($state eq 'result') {
        if (/^> *(.*)/) {
            push @result, "$1\n";
        } else {
            runTest($command, @result);
            if (/^\s*$/) {
                $state = 'pre';
                $command = "";
            } else {
                $state = 'command';
                $firstLine = $.;
                $command = $_;
            }
            @result = ();
        }
    }
}
if ($state eq 'result') {
    runTest($command, @result);
} elsif ($state eq 'command') {
    print "ERROR: testsuite error? last command has no result\n";
}
close Reader;
close Writer;
print "$numTests tests successful.\n";

sub runTest {
    my ($in, @out) = @_;
    print Writer $in;
    my @read;
    while (@read < @out) {
        push @read, scalar(<Reader>);
    }
    my $read = join('',@read);
    my $out  = join('',@out);
    if ($read ne $out) {
        print "ERROR: line $firstLine: expected:\n$out\ngot:\n$read\n";
        exit(1);
    }
    ++$numTests;
}
